/***************************************************************************************

***************************************************************************************/



/*
*************************************************************************************
* File name: defines.h
* Function: Headerfile containing some useful global definitions
*
*************************************************************************************
*/


#ifndef _DEFINES_H_
#define _DEFINES_H_

#include <stdlib.h>  

//for ivc profile
#define IVC_PROFILE
#define BASELINE_PROFILE_ID    0x20
#define P_SUB_TYPE

#define DCT_TRANS

#define TOTRUN_NUM    15

#define B_DIRECT_REMOVE

//--- block types for AEC ----
#define LUMA_8x8        2
#define CHROMA_DC       6
#define CHROMA_AC       7
#define NUM_BLOCK_TYPES 8

#define MAX_CODED_FRAME_SIZE 1500000         //!< bytes for one frame

#define TRACE           0                   //!< 0:Trace off 1:Trace on


#define MIN(a,b)  (((a)<(b)) ? (a) : (b))
#define MAX(a,b)  (((a)<(b)) ? (b) : (a))

#define absm(A) ((A)<(0) ? (-(A)):(A))      //!< abs macro, faster than procedure
#define MAX_VALUE       999999              //!< used for start value for some variables
#define Clip1(a)            ((a)>255?255:((a)<0?0:(a)))
#define Clip3(min,max,val) (((val)<(min))?(min):(((val)>(max))?(max):(val)))

// FLAGS and DEFINES.
#define B8_SIZE         8       // block size of transformed block

#define P8x8    8
#define I8MB    9
#define I4MB   10
#define IBLOCK  11
#define SI4MB   12
#define MAXMODE 13

#define IS_INTRA(MB)    ((MB)->mb_type==I8MB)
#define IS_INTER(MB)    ((MB)->mb_type!=I8MB)
#define IS_DIRECT(MB)   ((MB)->mb_type==0   && (img->type ==    B_IMG ))
#define IS_COPY(MB)     ((MB)->mb_type==0   && (img->type == P_IMG))
#define IS_P8x8(MB)     ((MB)->mb_type==P8x8)

// Quantization parameter range
#define MIN_QP          0
#define MAX_QP          63

#define BLOCK_SIZE      4
#define B8_SIZE    8
#define MB_BLOCK_SIZE   16

/* 8x8 intra prediction modes */
#define DC_PRED    2

// 8x8 chroma intra prediction modes
#define DC_PRED_8       0

#define EOS             1     //!< End Of Sequence
#define SOP             2     //!< Start Of Picture

#define DECODING_OK     0
#define SEARCH_SYNC     1


#ifndef WIN32
#define max(a, b)      ((a) > (b) ? (a) : (b))  //!< Macro returning max value
#define min(a, b)      ((a) < (b) ? (a) : (b))  //!< Macro returning min value
#endif
#define mmax(a, b)      ((a) > (b) ? (a) : (b)) //!< Macro returning max value
#define mmin(a, b)      ((a) < (b) ? (a) : (b)) //!< Macro returning min value
#define clamp(a,b,c) ( (a)<(b) ? (b) : ((a)>(c)?(c):(a)) )    //!< clamp a to the range of [b;c]

// mv prediction
#define MVPRED_xy_MIN   0
#define MVPRED_L        1
#define MVPRED_U        2
#define MVPRED_UR       3
// mv prediction end

#define DECODE_MB       1

#define BLOCK_MULTIPLE      (MB_BLOCK_SIZE/(BLOCK_SIZE*2))

#endif

