/***************************************************************************************

***************************************************************************************/

/*
*************************************************************************************
* File name: block.c
* Function: Description
*
*************************************************************************************
*/



#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <memory.h>

#include "defines.h"
#include "global.h"
#include "elements.h"
#include "block.h"
#include "vlc.h"
#include "golomb_dec.h"
#include "checksyntax.h"
#include "AEC.h" 
#define EP (edgepixels+20)

#ifdef WIN32
#define int16 __int16
#else
#define int16 int16_t
#endif

unsigned short IQ_TAB[64] = {

  32768,36061,38968,42495,46341,50535,55437,60424,
  32932,35734,38968,42495,46177,50535,55109,59933,
  65535,35734,38968,42577,46341,50617,55027,60097,
  32809,35734,38968,42454,46382,50576,55109,60056,
  65535,35734,38968,42495,46320,50515,55109,60076,
  65535,35744,38968,42495,46341,50535,55099,60087,
  65535,35734,38973,42500,46341,50535,55109,60097,
  32771,35734,38965,42497,46341,50535,55109,60099

};

short IQ_SHIFT[64] = {
  15,15,15,15,15,15,15,15,
  14,14,14,14,14,14,14,14,
  14,13,13,13,13,13,13,13,
  12,12,12,12,12,12,12,12,
  12,11,11,11,11,11,11,11,
  11,10,10,10,10,10,10,10,
  10,9,9,9,9,9,9,9,
  8,8,8,8,8,8,8,8

};

/*
*************************************************************************
* Function:Inverse transform 8x8 block.
* Input:
* Output:
* Return: 
* Attention:input curr_blk has to be [lines][pixels]
*************************************************************************
*/
#ifdef DCT_TRANS
void inv_transform_B8( int curr_blk[B8_SIZE][B8_SIZE] )
{
	int i, j ;
	int temp1[8], temp2[8];

	for( i=0; i<8; i++ )
	{
		// step 1
		temp1[0] = ( curr_blk[i][0] + curr_blk[i][4] )*181>>7 ;
		temp1[1] = ( curr_blk[i][0] - curr_blk[i][4] )*181>>7 ;
		temp1[2] = ( curr_blk[i][2]*196>>8 ) - ( curr_blk[i][6]*473>>8 ) ;
		temp1[3] = ( curr_blk[i][2]*473>>8 ) + ( curr_blk[i][6]*196>>8 ) ;

		temp2[4] = curr_blk[i][1] - curr_blk[i][7] ;
		temp2[7] = curr_blk[i][1] + curr_blk[i][7] ;
		temp2[5] = curr_blk[i][3]*181>>7 ;
		temp2[6] = curr_blk[i][5]*181>>7 ;
		temp1[4] = temp2[4] + temp2[6] ;
		temp1[5] = temp2[7] - temp2[5] ;
		temp1[6] = temp2[4] - temp2[6] ;
		temp1[7] = temp2[7] + temp2[5] ;

		// step 2
		temp2[0] = temp1[0] + temp1[3] ;
		temp2[3] = temp1[0] - temp1[3] ;
		temp2[1] = temp1[1] + temp1[2] ;
		temp2[2] = temp1[1] - temp1[2] ;
		temp2[4] = ( temp1[4]*301>>8 ) - ( temp1[7]*201>>8 ) ;
		temp2[7] = ( temp1[4]*201>>8 ) + ( temp1[7]*301>>8 ) ;
		temp2[5] = ( temp1[5]*710>>9 ) - ( temp1[6]*141>>9 ) ;
		temp2[6] = ( temp1[5]*141>>9 ) + ( temp1[6]*710>>9 ) ;

		// step 3
		curr_blk[i][0] = temp2[0] + temp2[7] ;
		curr_blk[i][7] = temp2[0] - temp2[7] ;
		curr_blk[i][1] = temp2[1] + temp2[6] ;
		curr_blk[i][6] = temp2[1] - temp2[6] ;
		curr_blk[i][2] = temp2[2] + temp2[5] ;
		curr_blk[i][5] = temp2[2] - temp2[5] ;
		curr_blk[i][3] = temp2[3] + temp2[4] ;
		curr_blk[i][4] = temp2[3] - temp2[4] ;
	}
	for( i=0; i<8; i++ )
	{
		// step 1
		temp1[0] = ( curr_blk[0][i] + curr_blk[4][i] )*181>>7 ;
		temp1[1] = ( curr_blk[0][i] - curr_blk[4][i] )*181>>7 ;
		temp1[2] = ( curr_blk[2][i]*196>>8 ) - ( curr_blk[6][i]*473>>8 ) ;
		temp1[3] = ( curr_blk[2][i]*473>>8 ) + ( curr_blk[6][i]*196>>8 ) ;

		temp2[4] = curr_blk[1][i] - curr_blk[7][i] ;
		temp2[7] = curr_blk[1][i] + curr_blk[7][i] ;
		temp2[5] = curr_blk[3][i]*181>>7 ;
		temp2[6] = curr_blk[5][i]*181>>7 ;
		temp1[4] = temp2[4] + temp2[6] ;
		temp1[5] = temp2[7] - temp2[5] ;
		temp1[6] = temp2[4] - temp2[6] ;
		temp1[7] = temp2[7] + temp2[5] ;

		// step 2
		temp2[0] = temp1[0] + temp1[3] ;
		temp2[3] = temp1[0] - temp1[3] ;
		temp2[1] = temp1[1] + temp1[2] ;
		temp2[2] = temp1[1] - temp1[2] ;
		temp2[4] = ( temp1[4]*301>>8 ) - ( temp1[7]*201>>8 ) ;
		temp2[7] = ( temp1[4]*201>>8 ) + ( temp1[7]*301>>8 ) ;
		temp2[5] = ( temp1[5]*710>>9 ) - ( temp1[6]*141>>9 ) ;
		temp2[6] = ( temp1[5]*141>>9 ) + ( temp1[6]*710>>9 ) ;

		// step 3
		curr_blk[0][i] = temp2[0] + temp2[7] ;
		curr_blk[7][i] = temp2[0] - temp2[7] ;
		curr_blk[1][i] = temp2[1] + temp2[6] ;
		curr_blk[6][i] = temp2[1] - temp2[6] ;
		curr_blk[2][i] = temp2[2] + temp2[5] ;
		curr_blk[5][i] = temp2[2] - temp2[5] ;
		curr_blk[3][i] = temp2[3] + temp2[4] ;
		curr_blk[4][i] = temp2[3] - temp2[4] ;

		for( j=0; j<8; j++ )
		{
			if( curr_blk[j][i]>0 )
			{
				curr_blk[j][i] = ( curr_blk[j][i] + 16 )>>5 ;
			}
			else
			{
				curr_blk[j][i] = -( ( -curr_blk[j][i] + 16 )>>5 ) ;
			}
		}
	}

}
#else
void inv_transform_B8 (int curr_blk[B8_SIZE][B8_SIZE])  // block to be inverse transformed.
{
	int xx, yy;
	int tmp[8];
	int t;
	int b[8];
	int clip1, clip2; 
	clip1=0-(1<<(8+7));
	clip2=(1<<(8+7))-1;
	for(yy=0; yy<8; yy++)
	{
		// Horizontal inverse transform
		// Reorder
		tmp[0]=curr_blk[yy][0];
		tmp[1]=curr_blk[yy][4];
		tmp[2]=curr_blk[yy][2];
		tmp[3]=curr_blk[yy][6];
		tmp[4]=curr_blk[yy][1];
		tmp[5]=curr_blk[yy][3];
		tmp[6]=curr_blk[yy][5];
		tmp[7]=curr_blk[yy][7];

		// Downleft Butterfly
		b[0] = ((tmp[4] - tmp[7])<<1) + tmp[4];
		b[1] = ((tmp[5] + tmp[6])<<1) + tmp[5];
		b[2] = ((tmp[5] - tmp[6])<<1) - tmp[6];
		b[3] = ((tmp[4] + tmp[7])<<1) + tmp[7];

		b[4] = ((b[0] + b[1] + b[3])<<1) + b[1];
		b[5] = ((b[0] - b[1] + b[2])<<1) + b[0];
		b[6] = ((-b[1] - b[2] + b[3])<<1) + b[3];
		b[7] = ((b[0] - b[2] - b[3])<<1) - b[2];

		// Upleft Butterfly
		t=((tmp[2]*10)+(tmp[3]<<2));
		tmp[3]=((tmp[2]<<2)-(tmp[3]*10));
		tmp[2]=t;

		t=(tmp[0]+tmp[1])<<3;
		tmp[1]=(tmp[0]-tmp[1])<<3;
		tmp[0]=t;

		b[0]=tmp[0]+tmp[2];
		b[1]=tmp[1]+tmp[3];
		b[2]=tmp[1]-tmp[3];
		b[3]=tmp[0]-tmp[2];	 

		curr_blk[yy][0]=(int)((Clip3(clip1,clip2,((b[0]+b[4])+(1<<2))))>>3); 
		curr_blk[yy][1]=(int)((Clip3(clip1,clip2,((b[1]+b[5])+(1<<2))))>>3);
		curr_blk[yy][2]=(int)((Clip3(clip1,clip2,((b[2]+b[6])+(1<<2))))>>3);
		curr_blk[yy][3]=(int)((Clip3(clip1,clip2,((b[3]+b[7])+(1<<2))))>>3);
		curr_blk[yy][7]=(int)((Clip3(clip1,clip2,((b[0]-b[4])+(1<<2))))>>3);
		curr_blk[yy][6]=(int)((Clip3(clip1,clip2,((b[1]-b[5])+(1<<2))))>>3);
		curr_blk[yy][5]=(int)((Clip3(clip1,clip2,((b[2]-b[6])+(1<<2))))>>3);
		curr_blk[yy][4]=(int)((Clip3(clip1,clip2,((b[3]-b[7])+(1<<2))))>>3);

	}
	// Vertical inverse transform
	for(xx=0; xx<8; xx++)
	{

		// Reorder
		tmp[0]=curr_blk[0][xx];
		tmp[1]=curr_blk[4][xx];
		tmp[2]=curr_blk[2][xx];
		tmp[3]=curr_blk[6][xx];
		tmp[4]=curr_blk[1][xx];
		tmp[5]=curr_blk[3][xx];
		tmp[6]=curr_blk[5][xx];
		tmp[7]=curr_blk[7][xx];

		// Downleft Butterfly
		b[0] = ((tmp[4] - tmp[7])<<1) + tmp[4];
		b[1] = ((tmp[5] + tmp[6])<<1) + tmp[5];
		b[2] = ((tmp[5] - tmp[6])<<1) - tmp[6];
		b[3] = ((tmp[4] + tmp[7])<<1) + tmp[7];

		b[4] = ((b[0] + b[1] + b[3])<<1) + b[1];
		b[5] = ((b[0] - b[1] + b[2])<<1) + b[0];
		b[6] = ((-b[1] - b[2] + b[3])<<1) + b[3];
		b[7] = ((b[0] - b[2] - b[3])<<1) - b[2];

		// Upleft Butterfly
		t=((tmp[2]*10)+(tmp[3]<<2));
		tmp[3]=((tmp[2]<<2)-(tmp[3]*10));
		tmp[2]=t;

		t=(tmp[0]+tmp[1])<<3;
		tmp[1]=(tmp[0]-tmp[1])<<3;
		tmp[0]=t;

		b[0]=tmp[0]+tmp[2];
		b[1]=tmp[1]+tmp[3];
		b[2]=tmp[1]-tmp[3];
		b[3]=tmp[0]-tmp[2];

		curr_blk[0][xx]=(Clip3(clip1,clip2,(b[0]+b[4])+64))>>7;  
		curr_blk[1][xx]=(Clip3(clip1,clip2,(b[1]+b[5])+64))>>7;
		curr_blk[2][xx]=(Clip3(clip1,clip2,(b[2]+b[6])+64))>>7;
		curr_blk[3][xx]=(Clip3(clip1,clip2,(b[3]+b[7])+64))>>7;
		curr_blk[7][xx]=(Clip3(clip1,clip2,(b[0]-b[4])+64))>>7;
		curr_blk[6][xx]=(Clip3(clip1,clip2,(b[1]-b[5])+64))>>7;
		curr_blk[5][xx]=(Clip3(clip1,clip2,(b[2]-b[6])+64))>>7;
		curr_blk[4][xx]=(Clip3(clip1,clip2,(b[3]-b[7])+64))>>7;

	}
}
#endif

void idct_dequant_B8(int block8x8,
                     int qp,                         // Quantization parameter
                     int curr_blk[B8_SIZE][B8_SIZE],
struct img_par *img
  )
{
  int  xx,yy;
  int  b8_y       = (block8x8 / 2) << 3;
  int  b8_x       = (block8x8 % 2) << 3;
  int  curr_val;
  int b8_cy;

  // inverse transform
  inv_transform_B8(curr_blk);

  if   (block8x8>5)    
    b8_cy=8; 
  else  
    b8_cy=0;      

  // normalize
  for(yy=0;yy<8;yy++)
    for(xx=0;xx<8;xx++)
    {
      if(block8x8 <= 3)
        curr_val = img->mpr[b8_x+xx][b8_y+yy] + curr_blk[yy][xx];
      else
        curr_val = img->mpr[xx][yy+b8_cy] + curr_blk[yy][xx]; 

      img->m7[xx][yy] = curr_blk[yy][xx] = clamp(curr_val,0,255);

      if(block8x8 <= 3)
        imgY[img->pix_y+b8_y+yy][img->pix_x+b8_x+xx] = (unsigned char)curr_blk[yy][xx];
      else
        imgUV[(block8x8-4)%2][img->pix_c_y+yy+b8_cy][img->pix_c_x+xx] = (unsigned char)curr_blk[yy][xx];    

    }

}

/*
*************************************************************************
* Function:
* Input:
* Output:
* Return: 
* Attention:
*************************************************************************
*/
void reconstruct8x8(int curr_blk[8][8],int block8x8)
{
  int yy,xx,block_y, block_x;
  int incr_y=1,off_y=0; 
  Macroblock *currMB   = &mb_data[img->current_mb_nr];//GB current_mb_nr];

  if(block8x8<4)
  {
    block_y = (block8x8/2)*8;
    block_x = (block8x8%2)*8;

    for(yy=0;yy<8;yy++)
    {
      for(xx=0;xx<8;xx++)
      {				
        imgY[img->pix_y+block_y+incr_y*yy+off_y][img->pix_x+block_x+xx] = curr_blk[yy][xx];
      }
    }
  }
  else
  {	
    for(yy=0;yy<8;yy++)
    {
      for(xx=0;xx<8;xx++)			
      {
        imgUV[block8x8-4][img->pix_c_y+yy][img->pix_c_x+xx] = curr_blk[yy][xx];
      }
    }
  }
}

/*
*************************************************************************
* Function:Copy region of img->m7 corresponding to block8x8 to curr_blk[][].
* Input:
* Output:
* Return: 
* Attention:img->m7 is [x][y] and curr_blk is [y][x].
*************************************************************************
*/

void get_curr_blk( int block8x8,struct img_par *img, int curr_blk[B8_SIZE][B8_SIZE])
{
  int  xx, yy;
  int  mb_y       = (block8x8 / 2) << 3;
  int  mb_x       = (block8x8 % 2) << 3;
  Macroblock *currMB   = &mb_data[img->current_mb_nr];/*lgp*/  

  for (yy=0; yy<B8_SIZE; yy++)
    for (xx=0; xx<B8_SIZE; xx++)
    {
      if (block8x8<=3)
        curr_blk[yy][xx] = img->m7[mb_x+xx][mb_y+yy];
      else
        curr_blk[yy][xx] = img->m8[block8x8-4][xx][yy];
    }
}

/*
*************************************************************************
* Function:Make Intra prediction for all 5 modes for 8*8 blocks.
bs_x and bs_y may be only 4 and 8.
img_x and img_y are pixels offsets in the picture.
* Input:
* Output:
* Return: 
* Attention:
*************************************************************************
*/
int intrapred(struct img_par *img,int img_x,int img_y)
{
  unsigned char edgepixels[40];
  int x,y,last_pix,new_pix;
  unsigned int x_off,y_off;
  int i,predmode;
  int b4_x,b4_y;
  int b4,b8,mb;
  int block_available_up,block_available_up_right;
  int block_available_left,block_available_left_down;
  int bs_x,bs_y;
  int b8_x,b8_y;
  Macroblock *currMB   = &mb_data[img->current_mb_nr];
  int MBRowSize = img->width / MB_BLOCK_SIZE;
  int off_up=1,incr_y=1,off_y=0;
  const int mb_nr    = img->current_mb_nr;   
  int mb_left_available = (img_x >= MB_BLOCK_SIZE)?currMB->slice_nr == mb_data[mb_nr-1].slice_nr:0; 
  int mb_up_available = (img_y >= MB_BLOCK_SIZE)?currMB->slice_nr == mb_data[mb_nr-MBRowSize].slice_nr:0; 
  int mb_up_right_available; 
  int mb_left_down_available;
  int mb_up_left_available;   
  int block_available_up_left; 

    if((img->mb_y==0)||(img->mb_x==img->width/MB_BLOCK_SIZE-1))
      mb_up_right_available =1;
    else if((img_y-img->pix_y)>0)
      mb_up_right_available =(img_x-img->pix_x)>0? (currMB->slice_nr == mb_data[mb_nr+1].slice_nr):1;  
    else
      mb_up_right_available =((img_x-img->pix_x)>0? (currMB->slice_nr == mb_data[mb_nr-MBRowSize+1].slice_nr):(currMB->slice_nr== mb_data[mb_nr-MBRowSize].slice_nr));  

    if((img->mb_x==0)||(img->mb_y==img->height/MB_BLOCK_SIZE-1))
      mb_left_down_available = 1;
    else if(img_x-img->pix_x>0)
      mb_left_down_available =(img_y-img->pix_y)>0? (currMB->slice_nr == mb_data[mb_nr+MBRowSize].slice_nr):1; 
    else
      mb_left_down_available =((img_y-img->pix_y)>0? (currMB->slice_nr == mb_data[mb_nr+MBRowSize-1].slice_nr):(currMB->slice_nr == mb_data[mb_nr-1].slice_nr)); 

  b8_x=img_x>>3;
  b8_y=img_y>>3;

  bs_x =  bs_y = 8;
  x_off=img_x&0x0FU;
  y_off=img_y&0x0FU;

  predmode = img->ipredmode[img_x/(2*BLOCK_SIZE)+1][img_y/(2*BLOCK_SIZE)+1];
  assert(predmode>=0);
  {
    b4_x=img_x>>2;
    b4_y=img_y>>2;
    mb=(b4_x>>2)+(b4_y>>2)*(img->width>>2);  b8=((b4_x&2)>>1)|(b4_y&2);  b4=(b4_x&1)|((b4_y&1)<<1);

    //check block up
    block_available_up=( b8_y-1>=0 && (mb_up_available||(img_y/8)%2));

    //check block up right
    block_available_up_right=( b8_x+1<(img->width>>3) && b8_y-1>=0 &&  mb_up_right_available);

    //check block left
    block_available_left=( b8_x - 1 >=0 && ( mb_left_available)||((img_x/8)%2));

    //check block left down
    block_available_left_down=( b8_x - 1>=0 && b8_y + 1 < (img->height>>3) &&  mb_left_down_available);

    if( ((img_x/8)%2)  && ((img_y/8)%2))
      block_available_up_right=0;

    if( ((img_x/8)%2)  || ((img_y/8)%2))
      block_available_left_down=0;

    //check block up left, 
    block_available_up_left= (b8_y-1>=0 && b8_x - 1>=0 && mb_up_left_available); 

    //get prediciton pixels
    if(block_available_up)
    {
      for(x=0;x<bs_x;x++)
        EP[x+1]=imgY[img_y-1][img_x+x];
      if(block_available_up_right)
      {
        for(x=0;x<bs_x;x++)
          EP[1+x+bs_x]=imgY[img_y-1][img_x+bs_x+x];
        for(;x<bs_y;x++)
          EP[1+x+bs_x]=imgY[img_y-1][img_x+bs_x+bs_x-1];
      }else
        for(x=0;x<bs_y;x++)
          EP[1+x+bs_x]=EP[bs_x];
      for(;x<bs_y+2;x++)
        EP[1+x+bs_x]=EP[bs_x+x];
      EP[0]=imgY[img_y-1][img_x];
    }
    if(block_available_left)
    {
      for(y=0;y<bs_y;y++)
        EP[-1-y]=imgY[img_y+y][img_x-1];
      if(block_available_left_down)
      {
        for(y=0;y<bs_y;y++)
          EP[-1-y-bs_y]=imgY[img_y+bs_y+y][img_x-1];
        for(;y<bs_x;y++)
          EP[-1-y-bs_y]=imgY[img_y+bs_y+bs_y-1][img_x-1];
      }
      else
        for(y=0;y<bs_x;y++)
          EP[-1-y-bs_y]=EP[-bs_y];

      for(;y<bs_x+2;y++)
        EP[-1-y-bs_y]=EP[-y-bs_y];

      EP[0]=imgY[img_y][img_x-1];
    }

      if(block_available_up&&block_available_left)
        EP[0]=imgY[img_y-1][img_x-1];
    
    //lowpass (Those emlements that are not needed will not disturb)
    last_pix=EP[-(bs_x+bs_y)];
    for(i=-(bs_x+bs_y);i<=(bs_x+bs_y);i++)
    {
      new_pix=( last_pix + (EP[i]<<1) + EP[i+1] + 2 )>>2;
      last_pix=EP[i];
      EP[i]=(unsigned char)new_pix;
    }
  }

  switch(predmode)
  {
  case DC_PRED:// 0 DC
    if(!block_available_up && !block_available_left)
      for(y=0UL;y<bs_y;y++)
        for(x=0UL;x<bs_x;x++)
          img->mpr[x+x_off][y+y_off]=128UL;

    if(block_available_up && !block_available_left)
      for(y=0UL;y<bs_y;y++)
        for(x=0UL;x<bs_x;x++)
          img->mpr[x+x_off][y+y_off]=EP[1+x];

    if(!block_available_up && block_available_left)
      for(y=0UL;y<bs_y;y++)
        for(x=0UL;x<bs_x;x++)
          img->mpr[x+x_off][y+y_off]=EP[-1-y];

    if(block_available_up && block_available_left)
      for(y=0UL;y<bs_y;y++)
        for(x=0UL;x<bs_x;x++)
          img->mpr[x+x_off][y+y_off]=(EP[1+x]+EP[-1-y])>>1;

    break;
  }//end switch predmode

  return DECODING_OK;
}

