/***************************************************************************************

***************************************************************************************/



/*
*************************************************************************************
* File name: 
* Function: 
*
*************************************************************************************
*/
#include "contributors.h"

#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <sys/timeb.h>
#include <string.h>
#include <memory.h>
#include <assert.h>
#ifdef WIN32
#include <IO.H>
#endif

#include "global.h"
#include "image.h"
#include "refbuf.h"
#include "mbuffer.h"
#include "header.h"
#include "memalloc.h"
#include "bitstream.h"
#include "vlc.h"
#include "configfile.h"
#include "AEC.h" 
#include "wquant.h"

#define TO_SAVE 4711
#define FROM_SAVE 4712
#define Clip(min,max,val) (((val)<(min))?(min):(((val)>(max))?(max):(val)))

static void code_a_picture(Picture *frame);
static void ReadOneFrame (int FrameNoInFile, int HeaderSize, int xs, int ys);
static void write_reconstructed_image();
static int  writeout_picture();
static int  writeout_slice();
static void find_snr();
static void frame_mode_buffer (int bit_frame, float snr_frame_y, float snr_frame_u, float snr_frame_v);
static void init_frame();


int terminate_picture();


void put_buffer_frame();
void interpolate_frame_to_fb();

static void CopyFrameToOrigYUVBuffers ();
static void UnifiedOneForthPix (pel_t ** imgY, pel_t ** imgU, pel_t ** imgV,
                                pel_t ** out4Y);
static void ReportFirstframe(int tmp_time);
static void ReportIntra(int tmp_time);
static void ReportP(int tmp_time);
static void ReportB(int tmp_time);

static int CalculateFrameNumber();  // Calculates the next frame number
static int FrameNumberInFile;       // The current frame number in the input file

extern void DecideMvRange();	

#define  IClip( Min, Max, Val) (((Val)<(Min))? (Min):(((Val)>(Max))? (Max):(Val)))

/*
*************************************************************************
* Function:
* Input:
* Output:
* Return: 
* Attention:
*************************************************************************
*/

void stuffing_byte(int n)
{
  int i;
  Bitstream *currStream;

  currStream = currBitStream;

  for(i=0; i<n; i++)
  {
    currStream->streamBuffer[currStream->byte_pos++] = 0x80;
    currStream->bits_to_go = 8;
    currStream->byte_buf = 0;
  }
}

/*
*************************************************************************
* Function:
* Input:
* Output:
* Return: 
* Attention:
*************************************************************************
*/

static void picture_header()
{
  int len=0;

  Bitstream *bitstream = currBitStream;
  img->cod_counter = 0;

  if(img->type==INTRA_IMG)
  { 
    len =len+IPictureHeader(img->number);
    img->count=img->count+2;         

  }
  else
    len = len+PBPictureHeader();



  // Update statistics
  stat->bit_slice += len;
  stat->bit_use_header[img->type] += len;
}


/*
*************************************************************************
* Function:Encodes one frame
* Input:
* Output:
* Return: 
* Attention:
*************************************************************************
*/

int encode_one_frame ()
{
  time_t ltime1;
  time_t ltime2;

#ifdef WIN32
  struct _timeb tstruct1;
  struct _timeb tstruct2;
#else
  struct timeb tstruct1;
  struct timeb tstruct2;
#endif

  int tmp_time;
  int bits_frm = 0;
  float dis_frm = 0, dis_frm_y = 0, dis_frm_u = 0, dis_frm_v = 0;

#ifdef WIN32
  _ftime (&tstruct1);           // start time ms
#else
  ftime (&tstruct1);
#endif
  time (&ltime1);               // start time s

  init_frame ();           // initial frame variables

  FrameNumberInFile = CalculateFrameNumber();

  ReadOneFrame (FrameNumberInFile, input->infile_header, input->img_width, input->img_height);  

  CopyFrameToOrigYUVBuffers();

  current_slice_bytepos=0;  
  
  //prepare buffer pionter for encoding process
  put_buffer_frame ();     //initialize frame buffer

  DecideMvRange();	
    if (img->type == B_IMG)
      Bframe_ctr++;         // Bframe_ctr only used for statistics, should go to stat->

  code_a_picture (frame_pic);

  //update the reference picture buffer if necessary
  if (img->type!=B_IMG)
      Update_Picture_Buffers();

  stat->bit_ctr_emulationprevention += stat->em_prev_bits_frm;	

  //interpolating the reconstructed frame if necessary
  frame_mode_buffer (bits_frm, dis_frm_y, dis_frm_u, dis_frm_v);
  
  writeout_picture();
  FreeBitstream();  

  find_snr ();
  free(pic_buf);

  time (&ltime2);               // end time sec
#ifdef WIN32
  _ftime (&tstruct2);           // end time ms
#else
  ftime (&tstruct2);            // end time ms
#endif

  tmp_time = (ltime2 * 1000 + tstruct2.millitm) - (ltime1 * 1000 + tstruct1.millitm);
  tot_time = tot_time + tmp_time;

  // Write reconstructed images
  if (input->output_enc_pic)  //output_enc_pic
    write_reconstructed_image ();

  if (IMG_NUMBER == 0)
    ReportFirstframe(tmp_time);
  else
  {
    switch (img->type)
    {
    case INTRA_IMG:
      //stat->bit_ctr_P += stat->bit_ctr - stat->bit_ctr_n; 
      stat->bit_ctr_0 += stat->bit_ctr - stat->bit_ctr_n;   
      if(input->seqheader_period!=0 && (img->number+1)%(input->seqheader_period*input->intra_period)==0  && input->successive_Bframe==0) {
        stat->bit_ctr_0 += 32;    
        stat->bit_ctr   += 32;
        stat->bit_use_stuffingBits[0] += 32;
      }
      ReportIntra(tmp_time);
      break;
    case B_IMG:
      stat->bit_ctr_B += stat->bit_ctr - stat->bit_ctr_n;
      if(input->seqheader_period!=0 && (img->number+1)%(input->seqheader_period*input->intra_period)==0 && img->b_frame_to_code==input->successive_Bframe) {
        stat->bit_ctr_B += 32;  
        stat->bit_ctr   += 32;
        stat->bit_use_stuffingBits[2] += 32;
      }
      ReportB(tmp_time);
      break;
    default:      // P, P_MULTPRED?
      stat->bit_ctr_P += stat->bit_ctr - stat->bit_ctr_n;
      if(input->seqheader_period!=0 && (img->number+1)%(input->seqheader_period*input->intra_period)==0  && input->successive_Bframe==0) {
        stat->bit_ctr_P += 32;   
        stat->bit_ctr   += 32;
        stat->bit_use_stuffingBits[1] += 32;
      }
      ReportP(tmp_time);
    }
  }

  stat->bit_ctr_n = stat->bit_ctr;

  if (IMG_NUMBER == 0)
    return 0;
  else
    return 1;
}

/*
*************************************************************************
* Function:This function write out a picture
* Input:
* Output:
* Return: 0 if OK,                                                         \n
1 in case of error
* Attention:
*************************************************************************
*/

static int writeout_picture()
{

  assert (currBitStream->bits_to_go == 8);    //! should always be the case, 
  //! the byte alignment is done in terminate_slice
  WriteBitstreamtoFile();

  return 0;   
}

/*
*************************************************************************
* Function:Encodes a frame picture
* Input:
* Output:
* Return: 
* Attention:
*************************************************************************
*/

static void code_a_picture (Picture *frame)
{
  stat->em_prev_bits_frm = 0;
  stat->em_prev_bits = &stat->em_prev_bits_frm;

  AllocateBitstream();


  picture_header();

#ifdef P_SUB_TYPE
  if(img->type==INTRA_IMG || img->type==INTER_IMG)
  img->p_sub_type_start=0;
#endif

  picture_data(frame);

  frame->bits_per_picture = 8 * (currBitStream->byte_pos);  

}

/*
*************************************************************************
* Function:Frame Mode Buffer
* Input:
* Output:
* Return: 
* Attention:
*************************************************************************
*/

static void frame_mode_buffer (int bit_frame, float snr_frame_y, float snr_frame_u, float snr_frame_v)
{   
#ifdef P_SUB_TYPE
  if (img->type != B_IMG && img->typeb<=L1_IMG)       //all I- and P-frames
    interpolate_frame_to_fb ();
#else
  if (img->type != B_IMG)       //all I- and P-frames
    interpolate_frame_to_fb ();
#endif
}

/*
*************************************************************************
* Function:Initializes the parameters for a new frame
* Input:
* Output:
* Return: 
* Attention:
*************************************************************************
*/
int prevP_no=-1;
int nextP_no=-1;
static void init_frame ()
{
  int i, j, k;
  int prevP_no, nextP_no;
   int widthMB, heightMB;

  img->current_mb_nr = 0;
  img->current_slice_nr = 0;

    widthMB = img->width  / MB_BLOCK_SIZE;
    heightMB = img->height / MB_BLOCK_SIZE;
    img->mb_no_currSliceLastMB = ( input->slice_row_nr != 0 )
      ? min(input->slice_row_nr * widthMB - 1, widthMB * heightMB - 1)
      : widthMB * heightMB - 1 ;
  
  stat->bit_slice = 0;
  img->coded_mb_nr = 0;

  img->mb_y = img->mb_x = 0;
  img->block_y = img->pix_y = img->pix_c_y = 0; 
  img->block_x = img->pix_x = img->block_c_x = img->pix_c_x = 0;

  refFrArr    = refFrArr_frm;
  fw_refFrArr = fw_refFrArr_frm;
  bw_refFrArr = bw_refFrArr_frm;

  if (img->type != B_IMG)
  {
    if(IMG_NUMBER==0)
      img->tr=0;
    else
    {
      img->tr = IMG_NUMBER * img->p_interval;
      img->tr = min(img->tr, input->no_frames - 1);
    }
    img->imgtr_last_prev_P_frm = img->imgtr_last_P_frm;

    if (img->type==INTRA_IMG && img->Seqheader_flag==1) {
      img->last_picture_distance = img->curr_picture_distance;
      img->curr_picture_distance = (IMG_NUMBER * (input->jumpd + 1))%256;
    }

    picture_distance = (IMG_NUMBER * (input->jumpd + 1))%256;   


    img->imgtr_last_P_frm = img->imgtr_next_P_frm;
    img->imgtr_next_P_frm = picture_distance;	

    if (IMG_NUMBER != 0 && input->successive_Bframe != 0)     // B pictures to encode
      nextP_tr_frm = picture_distance;                   

    //Rate control
    if(!input->RCEnable)
    {
#ifdef P_SUB_TYPE
		if (img->type == INTRA_IMG)
			img->qp = input->qp0;   // set quant. parameter for I-frame
		else
		{
			if(input->p_sub_type_coding && img->tag_hp)
			{
				if((img->number-intra_frame_number)%HPGOPSIZE==0)
					img->qp=input->qpN;
				else if((img->number-intra_frame_number)%HPGOPSIZE==HPGOPSIZE/2)
					img->qp=input->qpN+input->p_sub_type_delta0;
				else
					img->qp=input->qpN+input->p_sub_type_delta1;
			}
			else
				img->qp=input->qpN;
		}
#else
		if (img->type == INTRA_IMG)
			img->qp = input->qp0;   // set quant. parameter for I-frame
		else
			img->qp=input->qpN;
#endif
    }
  }
  else
  {
    prevP_no = (IMG_NUMBER - 1) * img->p_interval;
    nextP_no = (IMG_NUMBER) * img->p_interval;
    nextP_no         = min(nextP_no, input->no_frames - 1);

    img->b_interval =
      (int) ((float) (input->jumpd + 1) / (input->successive_Bframe + 1.0) + 0.49999);

    if(img->number-1!=0)
      img->tr = (IMG_NUMBER - 1) * (input->jumpd + 1) + img->b_interval * img->b_frame_to_code;      // from prev_P
    else
      img->tr = prevP_no + img->b_interval * img->b_frame_to_code;      // from prev_P

    picture_distance = ((IMG_NUMBER - 1) * (input->jumpd + 1) + img->b_interval * img->b_frame_to_code)%256; 

    if (img->tr >= nextP_no)
      img->tr = nextP_no - 1;

    //Rate control
    if(!input->RCEnable)
      img->qp = input->qpB;

    // initialize arrays
      for (k = 0; k < 2; k++)
        for (i = 0; i < img->height / BLOCK_SIZE; i++)
          for (j = 0; j < img->width / BLOCK_SIZE + 4; j++)
          {
            tmp_fwMV[k][i][j] = 0;
            tmp_bwMV[k][i][j] = 0;
            dfMV[k][i][j] = 0;
            dbMV[k][i][j] = 0;
          }

          for (i = 0; i < img->height / BLOCK_SIZE; i++)
            for (j = 0; j < img->width / BLOCK_SIZE; j++)
            {
              fw_refFrArr[i][j] = bw_refFrArr[i][j] = -1;
            }
  }

  img->total_number_mb = (img->width * img->height) / (MB_BLOCK_SIZE * MB_BLOCK_SIZE); 

}
/*
*************************************************************************
* Function:Writes reconstructed image(s) to file
This can be done more elegant!
* Input:
* Output:
* Return: 
* Attention:
*************************************************************************
*/

static void write_reconstructed_image ()
{
  int i, j, k;
  int start = 0, inc = 1;

#ifdef _BUGFIX_NON16MUL_PICSIZE
  int writeout_width=(img->width-img->auto_crop_right); 
  int writeout_height=(img->height-img->auto_crop_bottom);  
  int writeout_width_cr =(img->width-img->auto_crop_right)/2;  
  int writeout_height_cr=(img->height-img->auto_crop_bottom)/2; 
#endif

  unsigned char *buf;
  buf = malloc (writeout_width * writeout_height);

#ifdef WIN32
  if (p_dec != -1)
  {
    if (img->type != B_IMG)
    {
      // write reconstructed image (IPPP)
      //if (input->successive_Bframe == 0)
      if ((input->successive_Bframe == 0) || ((img->tr == input->no_frames - 1) && (input->no_frames % (input->successive_Bframe + 1) == 2)) )
      {
        for(j = 0; j < writeout_height; j++)
          memcpy(buf + j * writeout_width, &(imgY[j][0]), writeout_width);
        _write(p_dec, buf, writeout_height*writeout_width);

        for(j = 0; j < writeout_height_cr; j++)
          memcpy(buf + j * writeout_width_cr, &(imgUV[0][j][0]), writeout_width_cr);
        _write(p_dec, buf, writeout_height_cr*writeout_width_cr);

        for(j = 0; j < writeout_height_cr; j++)
          memcpy(buf + j * writeout_width_cr, &(imgUV[1][j][0]), writeout_width_cr);
        _write(p_dec, buf, writeout_height_cr*writeout_width_cr);
      }
      // write reconstructed image (IBPBP) : only intra written
      else if (IMG_NUMBER == 0 && input->successive_Bframe != 0)
      {
        for(j = 0; j < writeout_height; j++)
          memcpy(buf + j * writeout_width, &(imgY[j][0]), writeout_width);
        _write(p_dec, buf, writeout_height*writeout_width);

        for(j = 0; j < writeout_height_cr; j++)
          memcpy(buf + j * writeout_width_cr, &(imgUV[0][j][0]), writeout_width_cr);
        _write(p_dec, buf, writeout_height_cr*writeout_width_cr);

        for(j = 0; j < writeout_height_cr; j++)
          memcpy(buf + j * writeout_width_cr, &(imgUV[1][j][0]), writeout_width_cr);
        _write(p_dec, buf, writeout_height_cr*writeout_width_cr);
      }
      // next P picture. This is saved with recon B picture after B picture coding
      else if (IMG_NUMBER != 0 && input->successive_Bframe != 0)
      {
        for (i = start; i < writeout_height ; i += inc)
          for (j = 0; j < writeout_width ; j++)
            nextP_imgY[i][j] = imgY[i][j];

        for (k = 0; k < 2; ++k)
          for (i = start; i < writeout_height_cr; i += inc)
            for (j = 0; j < writeout_width_cr; j++)
              nextP_imgUV[k][i][j] = imgUV[k][i][j];
      }
    }
    else
    {
      for(j = 0; j < writeout_height; j++)
        memcpy(buf + j * writeout_width, &(imgY[j][0]), writeout_width);
      _write(p_dec, buf, writeout_height*writeout_width);

      for(j = 0; j < writeout_height_cr; j++)
        memcpy(buf + j * writeout_width_cr, &(imgUV[0][j][0]), writeout_width_cr);
      _write(p_dec, buf, writeout_height_cr*writeout_width_cr);

      for(j = 0; j < writeout_height_cr; j++)
        memcpy(buf + j * writeout_width_cr, &(imgUV[1][j][0]), writeout_width_cr);
      _write(p_dec, buf, writeout_height_cr*writeout_width_cr);

      //if (img->b_frame_to_code == input->successive_Bframe)
      if ( (img->b_frame_to_code == input->successive_Bframe) || (img->tr == input->no_frames - 2))
      {
        for(j = 0; j < writeout_height; j++)
          memcpy(buf + j * writeout_width, &(nextP_imgY[j][0]), writeout_width);
        _write(p_dec, buf, writeout_height*writeout_width);

        for(j = 0; j < writeout_height_cr; j++)
          memcpy(buf + j * writeout_width_cr, &(nextP_imgUV[0][j][0]), writeout_width_cr);
        _write(p_dec, buf, writeout_height_cr*writeout_width_cr);

        for(j = 0; j < writeout_height_cr; j++)
          memcpy(buf + j * writeout_width_cr, &(nextP_imgUV[1][j][0]), writeout_width_cr);
        _write(p_dec, buf, writeout_height_cr*writeout_width_cr);
      }
    }
  }

#else
  if (p_dec != NULL)
  {
    if (img->type != B_IMG)
    {
      // write reconstructed image (IPPP)
      if (input->successive_Bframe == 0)
      {
        for (i = start; i < writeout_height /*img->height*/; i += inc)
          for (j = 0; j < writeout_width /*img->width*/; j++)
            fputc (imgY[i][j], p_dec);

        for (k = 0; k < 2; ++k)
          for (i = start; i < writeout_height_cr; i += inc)
            for (j = 0; j < writeout_width_cr; j++)
              fputc (imgUV[k][i][j], p_dec);
      }
      // write reconstructed image (IBPBP) : only intra written
      else if (IMG_NUMBER == 0 && input->successive_Bframe != 0)
      {
        for (i = start; i < writeout_height /*img->height*/; i += inc)
          for (j = 0; j < writeout_width /*img->width*/; j++)
            fputc (imgY[i][j], p_dec);

        for (k = 0; k < 2; ++k)
          for (i = start; i < writeout_height_cr; i += inc)
            for (j = 0; j < writeout_width_cr; j++)
              fputc (imgUV[k][i][j], p_dec);
      }

      // next P picture. This is saved with recon B picture after B picture coding
      if (IMG_NUMBER != 0 && input->successive_Bframe != 0)
      {
        for (i = start; i < writeout_height /*img->height*/; i += inc)
          for (j = 0; j < writeout_width /*img->width*/; j++)
            nextP_imgY[i][j] = imgY[i][j];

        for (k = 0; k < 2; ++k)
          for (i = start; i < writeout_height_cr; i += inc)
            for (j = 0; j < writeout_width_cr; j++)
              nextP_imgUV[k][i][j] = imgUV[k][i][j];
      }
    }
    else
    {
      for (i = start; i < writeout_height /*img->height*/; i += inc)
        for (j = 0; j < writeout_width /*img->width*/; j++)
          fputc (imgY[i][j], p_dec);

      for (k = 0; k < 2; ++k)
        for (i = start; i < writeout_height_cr; i += inc)
          for (j = 0; j < writeout_width_cr; j++)
            fputc (imgUV[k][i][j], p_dec);

      // If this is last B frame also store P frame
      if (img->b_frame_to_code == input->successive_Bframe)
      {
        // save P picture
        for (i = start; i < writeout_height /*img->height*/; i += inc)
          for (j = 0; j < writeout_width /*img->width*/; j++)
            fputc (nextP_imgY[i][j], p_dec);

        for (k = 0; k < 2; ++k)
          for (i = start; i < writeout_height_cr; i += inc)
            for (j = 0; j < writeout_width_cr; j++)
              fputc (nextP_imgUV[k][i][j], p_dec);
      }
    }
  }
#endif  

  free (buf);
}

/*
*************************************************************************
* Function:Choose interpolation method depending on MV-resolution
* Input:
* Output:
* Return: 
* Attention:
*************************************************************************
*/
void interpolate_frame_to_fb ()
{                             // write to mref[]
  UnifiedOneForthPix (imgY, imgUV[0], imgUV[1], mref[0]);
}

/*
*************************************************************************
* Function:Upsample 4 times, store them in out4x.  Color is simply copied
* Input:srcy, srcu, srcv, out4y, out4u, out4v
* Output:
* Return: 
* Attention:Side Effects_
Uses (writes) img4Y_tmp.  This should be moved to a static variable
in this module
*************************************************************************
*/
#define  IClip( Min, Max, Val) (((Val)<(Min))? (Min):(((Val)>(Max))? (Max):(Val)))

static void UnifiedOneForthPix (pel_t ** imgY, pel_t ** imgU, pel_t ** imgV,
                                pel_t ** out4Y)
{
  int is;
  int i, j;// j4;
  int ie2, je2, jj, maxy;
  //new filter
  //horizontal 1/4, 1/2, 3/4 interpolation
  for (j = -IMG_PAD_SIZE; j < img->height + IMG_PAD_SIZE; j++)
		for (i = -IMG_PAD_SIZE; i < img->width + IMG_PAD_SIZE; i++)
		{
			jj = IClip (0, img->height - 1, j);

			//horizontal 1/2, filter coefficients: -1, 4, -11, 40, 40, -11, 4, -1
			is = 40 * (imgY[jj][IClip (0, img->width - 1, i)] + imgY[jj][IClip (0, img->width - 1, i + 1)]) +
				4  * (imgY[jj][IClip (0, img->width - 1, i - 2)] + imgY[jj][IClip (0, img->width - 1, i + 3)])
				- 11  * (imgY[jj][IClip (0, img->width - 1, i - 1)] + imgY[jj][IClip (0, img->width - 1, i + 2)])
				-       (imgY[jj][IClip (0, img->width - 1, i - 3)] + imgY[jj][IClip (0, img->width - 1, i + 4)]);
			//store horizontal 1/2 pixel value
			img4Y_tmp[(j + IMG_PAD_SIZE)*4][(i + IMG_PAD_SIZE) * 4 + 2] = is;

			//horizontal 1/4, filter coefficients: -1, 4, -10, 57, 18, -6, 3, -1
			is = (57 * imgY[jj][IClip (0, img->width - 1, i)]     + 18 * imgY[jj][IClip (0, img->width - 1, i + 1)] +
				4  * imgY[jj][IClip (0, img->width - 1, i - 2)] + 3  * imgY[jj][IClip (0, img->width - 1, i + 3)])
				-(10 * imgY[jj][IClip (0, img->width - 1, i - 1)] + 6  * imgY[jj][IClip (0, img->width - 1, i + 2)] +
				imgY[jj][IClip (0, img->width - 1, i - 3)] +      imgY[jj][IClip (0, img->width - 1, i + 4)]);
			//store horizontal 1/4 pixel value
			img4Y_tmp[(j + IMG_PAD_SIZE)*4][(i + IMG_PAD_SIZE) * 4 + 1] = is;

			//horizontal 3/4, filter coefficients: -1, 3, -6, 18, 57, -10, 4, -1
			is = (18 * imgY[jj][IClip (0, img->width - 1, i)]     + 57 * imgY[jj][IClip (0, img->width - 1, i + 1)] +
				3  * imgY[jj][IClip (0, img->width - 1, i - 2)] + 4  * imgY[jj][IClip (0, img->width - 1, i + 3)])
				-(6  * imgY[jj][IClip (0, img->width - 1, i - 1)] + 10 * imgY[jj][IClip (0, img->width - 1, i + 2)] +
				imgY[jj][IClip (0, img->width - 1, i - 3)] +      imgY[jj][IClip (0, img->width - 1, i + 4)]);
			//store horizontal 3/4 pixel value
			img4Y_tmp[(j + IMG_PAD_SIZE)*4][(i + IMG_PAD_SIZE) * 4 + 3] = is;

			//store 1/1 pixel value
			img4Y_tmp[(j + IMG_PAD_SIZE)*4][(i + IMG_PAD_SIZE) * 4] = imgY[IClip (0, img->height - 1, j)][IClip (0, img->width - 1, i)];
		}

		//vertical 1/4, 1/2, 3/4 interpolation
		for (i = 0; i < (img->width + 2 * IMG_PAD_SIZE) * 4; i+=4)
		{
			for (j = 0; j < (img->height + 2 * IMG_PAD_SIZE) * 4; j+=4)
			{
				maxy = (img->height + 2 * IMG_PAD_SIZE) * 4 - 4;

				//vertical 1/2, filter coefficients: -1, 4, -11, 40, 40, -11, 4, -1
				is = 40 * (img4Y_tmp[j][i]               + img4Y_tmp[min (maxy, j + 4)][i])+
					4  * (img4Y_tmp[max (0, j - 8)][i]  + img4Y_tmp[min (maxy, j + 12)][i])
					-  11 * (img4Y_tmp[max (0, j - 4)][i]  + img4Y_tmp[min (maxy, j + 8)][i])
					-       (img4Y_tmp[max (0, j - 12)][i] + img4Y_tmp[min (maxy, j + 16)][i]);
				img4Y_tmp[j + 2][i] =is;//store vertical 1/2 pixel value

				//vertical 1/4, filter coefficients; -1, 4, -10, 57, 18, -6, 3, -1
				is = (57 * img4Y_tmp[j][i]               + 18 * img4Y_tmp[min (maxy, j + 4)][i]+
					4 * img4Y_tmp[max (0, j - 8)][i]  +  3 * img4Y_tmp[min (maxy, j + 12)][i])
					-(10 * img4Y_tmp[max (0, j - 4)][i]  +  6 * img4Y_tmp[min (maxy, j + 8)][i]+
					img4Y_tmp[max (0, j - 12)][i] + img4Y_tmp[min (maxy, j + 16)][i]);
				img4Y_tmp[j + 1][i] =is;//store vertical 1/4 pixel value

				//vertical 3/4, filter coefficients: -1, 3, -6, 18, 57, -10, 4, -1
				is = (18 * img4Y_tmp[j][i]               + 57 * img4Y_tmp[min (maxy, j + 4)][i]+
					3 * img4Y_tmp[max (0, j - 8)][i]  +  4 * img4Y_tmp[min (maxy, j + 12)][i])
					-( 6 * img4Y_tmp[max (0, j - 4)][i]  + 10 * img4Y_tmp[min (maxy, j + 8)][i]+
					img4Y_tmp[max (0, j - 12)][i] + img4Y_tmp[min (maxy, j + 16)][i]);
				img4Y_tmp[j + 3][i] =is;//store vertical 3/4 pixel value


			}

		}

		//vertical interpolation, remaining 9 fractional-pel pixels
		for (i = 0; i < (img->width  + 2 * IMG_PAD_SIZE)*4; i++)
		{
			for (j = 0; j < (img->height  + 2 * IMG_PAD_SIZE)*4; j += 4)
			{
				// e,i,p; f,j,q; g,k,r
				if (i % 4 != 0)//namely (i % 4 == 1)||(i % 4 == 2)||(i % 4 == 3)
				{
					//e,or f, or g. vertical 1/4 interpolation, filter coefficients: 2, -9, 57, 17, -4, 1
					img4Y_tmp[j+1][i]=IClip(0,255,
						(int) ((2  * img4Y_tmp[IClip (0, (img->height + 2 * IMG_PAD_SIZE) * 4 - 4, j-8)][i] +
						57 * img4Y_tmp[IClip (0, (img->height + 2 * IMG_PAD_SIZE) * 4 - 4, j  )][i] +
						17 * img4Y_tmp[IClip (0, (img->height + 2 * IMG_PAD_SIZE) * 4 - 4, j+4)][i] +
						1  * img4Y_tmp[IClip (0, (img->height + 2 * IMG_PAD_SIZE) * 4 - 4, j+12)][i] -
						9  * img4Y_tmp[IClip (0, (img->height + 2 * IMG_PAD_SIZE) * 4 - 4, j-4)][i] -
						4  * img4Y_tmp[IClip (0, (img->height + 2 * IMG_PAD_SIZE) * 4 - 4, j+8)][i] + 2048) / 4096)
						);

					//i,or j, or k. vertical 1/2 interpolation, filter coefficients: 2, -9, 39, 39, -9, 2
					img4Y_tmp[j+2][i]=IClip(0,255,
						(int) ((2  * (img4Y_tmp[IClip (0, (img->height + 2 * IMG_PAD_SIZE) * 4 - 4, j-8)][i] +
						img4Y_tmp[IClip (0, (img->height + 2 * IMG_PAD_SIZE) * 4 - 4, j+12)][i]) +
						39 * (img4Y_tmp[IClip (0, (img->height + 2 * IMG_PAD_SIZE) * 4 - 4, j  )][i] + 
						img4Y_tmp[IClip (0, (img->height + 2 * IMG_PAD_SIZE) * 4 - 4, j+4)][i]) -
						9  * (img4Y_tmp[IClip (0, (img->height + 2 * IMG_PAD_SIZE) * 4 - 4, j-4)][i] +
						img4Y_tmp[IClip (0, (img->height + 2 * IMG_PAD_SIZE) * 4 - 4, j+8)][i]) + 2048) / 4096)
						);

					//p, or q, or r.vertical 3/4 interpolation, filter coefficients: 1, -4, 17, 57, -9, 2
					img4Y_tmp[j+3][i]=IClip(0,255,
						(int) ((1  * img4Y_tmp[IClip (0, (img->height + 2 * IMG_PAD_SIZE) * 4 - 4, j-8)][i] +
						17 * img4Y_tmp[IClip (0, (img->height + 2 * IMG_PAD_SIZE) * 4 - 4, j  )][i] +
						57 * img4Y_tmp[IClip (0, (img->height + 2 * IMG_PAD_SIZE) * 4 - 4, j+4)][i] +
						2  * img4Y_tmp[IClip (0, (img->height + 2 * IMG_PAD_SIZE) * 4 - 4, j+12)][i] -
						4  * img4Y_tmp[IClip (0, (img->height + 2 * IMG_PAD_SIZE) * 4 - 4, j-4)][i] -
						9  * img4Y_tmp[IClip (0, (img->height + 2 * IMG_PAD_SIZE) * 4 - 4, j+8)][i] + 2048) / 4096)
						);
				}
			}
		}


		for (j = 0; j < (img->height  + 2 * IMG_PAD_SIZE)*4; j ++)
			for (i = 0; i < (img->width  + 2 * IMG_PAD_SIZE)*4; i ++)
			{
				if (((i % 4 == 0)||(j % 4 == 0))&&(!((i % 4 == 0)&&(j % 4 == 0)))) //((i % 4 == 0)||(j % 4 == 0))//round a, b, c; d, h, n
					img4Y_tmp[j][i]=IClip(0, 255, (int) (img4Y_tmp[j][i] + 32) / 64);
			}


			for (j = 0; j < (img->height  + 2 * IMG_PAD_SIZE)*4; j ++)
				for (i = 0; i < (img->width  + 2 * IMG_PAD_SIZE)*4; i ++)
				{
					PutPel_14 (out4Y, j - IMG_PAD_SIZE * 4, i - IMG_PAD_SIZE * 4, (pel_t) img4Y_tmp[j][i]);
				}
}

/*
*************************************************************************
* Function:Find SNR for all three components
* Input:
* Output:
* Return: 
* Attention:
*************************************************************************
*/

static void find_snr ()
{
  int i, j;
  int diff_y, diff_u, diff_v;
  int impix;
  int uvformat = 4;  

  //  Calculate  PSNR for Y, U and V.
  //     Luma.
  impix = img->height * img->width;

  diff_y = 0;
  for (i = 0; i < img->width; ++i)
  {
    for (j = 0; j < img->height; ++j)
    {
      diff_y += img->quad[imgY_org[j][i] - imgY[j][i]];
    }
  }

  //     Chroma.
  diff_u = 0;
  diff_v = 0;

  for (i = 0; i < img->width_cr; i++)
  {
    for (j = 0; j < img->height_cr; j++)
    {
      diff_u += img->quad[imgUV_org[0][j][i] - imgUV[0][j][i]];
      diff_v += img->quad[imgUV_org[1][j][i] - imgUV[1][j][i]];
    }
  }

  //  Collecting SNR statistics
  if (diff_y != 0)
  {
    snr->snr_y = (float) (10 * log10 (65025 * (float) impix / (float) diff_y));         // luma snr for current frame
    snr->snr_u = (float) (10 * log10 (65025 * (float) impix / (float) (/*4*/uvformat * diff_u)));   // u croma snr for current frame, 1/4 of luma samples, 
    snr->snr_v = (float) (10 * log10 (65025 * (float) impix / (float) (/*4*/uvformat * diff_v)));   // v croma snr for current frame, 1/4 of luma samples, 
  }

  if (img->number == 0)
  {
    snr->snr_y1 = (float) (10 * log10 (65025 * (float) impix / (float) diff_y));        // keep luma snr for first frame
    snr->snr_u1 = (float) (10 * log10 (65025 * (float) impix / (float) (/*4*/uvformat * diff_u)));  // keep croma u snr for first frame
    snr->snr_v1 = (float) (10 * log10 (65025 * (float) impix / (float) (/*4*/uvformat * diff_v)));  // keep croma v snr for first frame
    snr->snr_ya = snr->snr_y1;
    snr->snr_ua = snr->snr_u1;
    snr->snr_va = snr->snr_v1;
  }
  // B pictures
  else
  {
    snr->snr_ya = (float) (snr->snr_ya * (img->number + Bframe_ctr) + snr->snr_y) / (img->number + Bframe_ctr + 1); // average snr lume for all frames inc. first
    snr->snr_ua = (float) (snr->snr_ua * (img->number + Bframe_ctr) + snr->snr_u) / (img->number + Bframe_ctr + 1); // average snr u croma for all frames inc. first
    snr->snr_va = (float) (snr->snr_va * (img->number + Bframe_ctr) + snr->snr_v) / (img->number + Bframe_ctr + 1); // average snr v croma for all frames inc. first
  }

}

/*
*************************************************************************
* Function:Find distortion for all three components
* Input:
* Output:
* Return: 
* Attention:
*************************************************************************
*/
void find_distortion ()
{
  int i, j;
  int diff_y, diff_u, diff_v;
  int impix;

  //  Calculate  PSNR for Y, U and V.
  //     Luma.
  impix = img->height * img->width;

  diff_y = 0;
  for (i = 0; i < img->width; ++i)
  {
    for (j = 0; j < img->height; ++j)
    {
      diff_y += img->quad[abs (imgY_org[j][i] - imgY[j][i])];
    }
  }

  //     Chroma.
  diff_u = 0;
  diff_v = 0;

  for (i = 0; i < img->width_cr; i++)
  {
    for (j = 0; j < img->height_cr; j++)
    {
      diff_u += img->quad[abs (imgUV_org[0][j][i] - imgUV[0][j][i])];
      diff_v += img->quad[abs (imgUV_org[1][j][i] - imgUV[1][j][i])];
    }
  }

  // Calculate real PSNR at find_snr_avg()
  snr->snr_y = (float) diff_y;
  snr->snr_u = (float) diff_u;
  snr->snr_v = (float) diff_v;

}

/*
*************************************************************************
* Function:
* Input:
* Output:
* Return: 
* Attention:
*************************************************************************
*/

static void ReportFirstframe(int tmp_time)
{
  

  FILE *file = fopen("stat.dat","at");

  fprintf(file,"\n -------------------- DEBUG_INFO_START -------------------- \n");

  fprintf (file,"%3d(I)  %8d %4d %7.4f %7.4f %7.4f  %5d          %3d      %3s\n",
    img->tr, stat->bit_ctr - stat->bit_ctr_n,
    img->qp, snr->snr_y, snr->snr_u, snr->snr_v, tmp_time,
    intras, "FRM");

  fclose(file);

  printf ("%3d(I)  %8d %4d %7.4f %7.4f %7.4f  %5d       %s \n",
    img->tr, stat->bit_ctr - stat->bit_ctr_n,
    img->qp, snr->snr_y, snr->snr_u, snr->snr_v, tmp_time,  "FRM" );

  stat->bitr0 = stat->bitr;
  stat->bit_ctr_0 = stat->bit_ctr;

}

/*
*************************************************************************
* Function:
* Input:
* Output:
* Return: 
* Attention:
*************************************************************************
*/

static void ReportIntra(int tmp_time)
{
  FILE *file = fopen("stat.dat","at");

  fprintf (file,"%3d(I)  %8d %4d %7.4f %7.4f %7.4f     %5d             \n",
    img->tr, stat->bit_ctr - stat->bit_ctr_n,
    img->qp, snr->snr_y, snr->snr_u, snr->snr_v, tmp_time );

  fclose(file);

  printf ("%3d(I)  %8d %4d %7.4f %7.4f %7.4f  %5d %3s\n",
    img->tr, stat->bit_ctr - stat->bit_ctr_n,
    img->qp, snr->snr_y, snr->snr_u, snr->snr_v, tmp_time, "FRM");

}

/*
*************************************************************************
* Function:
* Input:
* Output:
* Return: 
* Attention:
*************************************************************************
*/

static void ReportB(int tmp_time)
{
  FILE *file = fopen("stat.dat","at");

  fprintf (file,"%3d(B)  %8d %4d %7.4f %7.4f %7.4f  %5d        \n",
    img->tr, stat->bit_ctr - stat->bit_ctr_n, img->qp,
    snr->snr_y, snr->snr_u, snr->snr_v, tmp_time);

  fclose(file);

  printf ("%3d(B)  %8d %4d %7.4f %7.4f %7.4f  %5d       %3s\n",
    img->tr, stat->bit_ctr - stat->bit_ctr_n, img->qp,
    snr->snr_y, snr->snr_u, snr->snr_v, tmp_time, "FRM");

}

/*
*************************************************************************
* Function:
* Input:
* Output:
* Return: 
* Attention:
*************************************************************************
*/

static void ReportP(int tmp_time)
{            
  FILE *file = fopen("stat.dat","at");
  char typ = (img->type==INTRA_IMG)?'I':'P';

  fprintf (file,"%3d(%c)  %8d %4d %7.4f %7.4f %7.4f  %5d        %3d\n",
    img->tr, typ, stat->bit_ctr - stat->bit_ctr_n, img->qp, snr->snr_y,
    snr->snr_u, snr->snr_v, tmp_time,
    intras);

  fclose(file);

  printf ("%3d(%c)  %8d %4d %7.4f %7.4f %7.4f  %5d       %3s     %3d    \n",
    img->tr, typ, stat->bit_ctr - stat->bit_ctr_n, img->qp, snr->snr_y,
    snr->snr_u, snr->snr_v, tmp_time,  "FRM",intras);

}

/*
*************************************************************************
* Function:Copies contents of a Sourceframe structure into the separate YUV buffers represented by
*    variables imgY_org_frm and imgUV_org_frm.  No other side effects
* Input:  sf the source frame the frame is to be taken from
* Output:
* Return: 
* Attention:
*************************************************************************
*/
// image padding if necessary
static void CopyFrameToOrigYUVBuffers()
{
  int x, y;
  byte *u_buffer,*v_buffer;
  int input_height_cr = input->img_height/2; 
  int img_height_cr   = img->height/2;             
  int offset_units;

  u_buffer=imgY_org_buffer+input->img_width*input->img_height;
  v_buffer=imgY_org_buffer+input->img_width*input->img_height*5/4;

  // Y
  for (y=0; y<input->img_height; y++)
      for (x=0; x<input->img_width; x++)
			imgY_org_frm [y][x] = imgY_org_buffer[y*input->img_width+x];

  // UV
  for (y=0; y<input_height_cr; y++) 
      for (x=0; x<input->img_width/2; x++)
      {
        imgUV_org_frm[0][y][x] = u_buffer[y*input->img_width/2+x];
        imgUV_org_frm[1][y][x] = v_buffer[y*input->img_width/2+x];
      }

  // Y's right padding if necessary
  for (y=0; y<input->img_height; y++)
    for (x=input->img_width; x<img->width; x++)
      imgY_org_frm [y][x] = imgY_org_frm [y][x-1];

  //Y's padding bottom border if necessary
  for (y=input->img_height; y<img->height; y++)
    for (x=0; x<img->width; x++)
      imgY_org_frm [y][x] = imgY_org_frm [y-1][x];

  // UV's right padding if necessary
  for (y=0; y<input_height_cr; y++) 
    for (x=input->img_width/2; x<img->width/2; x++)
    {
      imgUV_org_frm [0][y][x] = imgUV_org_frm [0][y][x-1];
      imgUV_org_frm [1][y][x] = imgUV_org_frm [1][y][x-1];
    }

    // UV's bottom padding bottom if necessary
    for (y=input_height_cr; y<img_height_cr; y++) 
      for (x=0; x<img->width/2; x++)
      {
        imgUV_org_frm [0][y][x] = imgUV_org_frm [0][y-1][x];
        imgUV_org_frm [1][y][x] = imgUV_org_frm [1][y-1][x];
      }  
}

/*
*************************************************************************
* Function: Calculates the absolute frame number in the source file out
of various variables in img-> and input->
* Input:
* Output:
* Return: frame number in the file to be read
* Attention: \side effects
global variable frame_no updated -- dunno, for what this one is necessary
*************************************************************************
*/

static int CalculateFrameNumber()
{
  frame_no = img->tr;
  // printf the position of frame in input sequence
  //printf("%5d",img->tr);
  return img->tr;
}

/*
*************************************************************************
* Function:Reads one new frame from file
* Input: FrameNoInFile: Frame number in the source file
HeaderSize: Number of bytes in the source file to be skipped
xs: horizontal size of frame in pixels, must be divisible by 16
ys: vertical size of frame in pixels, must be divisible by 16 
sf: Sourceframe structure to which the frame is written
* Output:
* Return: 
* Attention:
*************************************************************************
*/

static void ReadOneFrame (int FrameNoInFile, int HeaderSize, int xs, int ys)
{
  int input_width_cr  = input->img_width/2;                        
  int input_height_cr = input->img_height/2;
  unsigned int  bytes_y = input->img_width * input->img_height;      
  const unsigned int  bytes_uv = input_width_cr * input_height_cr;     
#ifdef WIN32
  const __int64 framesize_in_bytes = bytes_y + 2*bytes_uv;
#else  
  const int framesize_in_bytes = bytes_y + 2*bytes_uv;
#endif  
  //int stuff_height_cr = (input->img_height-input->stuff_height)/2;
  int off_y  = input->img_width*input->img_height;
  int off_uv = input_width_cr * input_height_cr;                    

  assert (FrameNumberInFile == FrameNoInFile);

#ifdef WIN32
  if (_lseeki64 (p_in, framesize_in_bytes * FrameNoInFile + HeaderSize, SEEK_SET)==-1)
    error ("ReadOneFrame: cannot fseek to (Header size) in p_in", -1);

  if (_read (p_in, imgY_org_buffer, bytes_y) != (int)bytes_y)
  {
    printf ("ReadOneFrame: cannot read %d bytes from input file, unexpected EOF?, exiting", bytes_y);
    exit (-1);
  }

  if (_read (p_in, imgY_org_buffer + off_y, bytes_uv) != (int)bytes_uv)
  {
    printf ("ReadOneFrame: cannot read %d bytes from input file, unexpected EOF?, exiting", bytes_uv);
    exit (-1);
  }

  if (_read (p_in, imgY_org_buffer + off_y + off_uv, bytes_uv) != (int)bytes_uv)
  {
    printf ("ReadOneFrame: cannot read %d bytes from input file, unexpected EOF?, exiting", bytes_uv);
    exit (-1);
  }
#else
  if (fseek (p_in, framesize_in_bytes * FrameNoInFile + HeaderSize, SEEK_SET) != 0)
    error ("ReadOneFrame: cannot fseek to (Header size) in p_in", -1);

  if (fread (imgY_org_buffer, 1, bytes_y, p_in) != (unsigned )bytes_y)
  {
    printf ("ReadOneFrame: cannot read %d bytes from input file, unexpected EOF?, exiting", bytes_y);
    exit (-1);
  }

  if (fread (imgY_org_buffer + off_y, 1, bytes_uv, p_in) != (unsigned )bytes_uv)
  {
    printf ("ReadOneFrame: cannot read %d bytes from input file, unexpected EOF?, exiting", bytes_uv);
    exit (-1);
  }

  if (fread (imgY_org_buffer + off_y + off_uv, 1, bytes_uv, p_in) != (unsigned )bytes_uv)
  {
    printf ("ReadOneFrame: cannot read %d bytes from input file, unexpected EOF?, exiting", bytes_uv);
    exit (-1);
  }
#endif
}

/*
*************************************************************************
* Function:point to frame coding variables 
* Input:
* Output:
* Return: 
* Attention:
*************************************************************************
*/

void put_buffer_frame()
{
  int i,j;

  imgY_org  = imgY_org_frm;
  imgUV_org = imgUV_org_frm;  
  tmp_mv    = tmp_mv_frm;

  //initialize ref index 1/4 pixel
  for(i=0;i<2;i++)
  {
    mref[i] = mref_frm[i];
  }

  //integer pixel for chroma
  for(i=0;i<2;i++)
    for(j=0;j<2;j++)
    {
      mcef[i][j]   = ref_frm[i][j+1];  
    }

    //integer pixel for luma
    for(i=0;i<2;i++)
      Refbuf11[i] = &ref_frm[i][0][0][0];

    //current reconstructed image 

    imgY = imgY_frm = current_frame[0];
    imgUV = imgUV_frm =  &current_frame[1];

    refFrArr    = refFrArr_frm;
    fw_refFrArr = fw_refFrArr_frm;
    bw_refFrArr = bw_refFrArr_frm;
}

/*
*************************************************************************
* Function:update the decoder picture buffer
* Input:frame number in the bitstream and the video sequence
* Output:
* Return: 
* Attention:
*************************************************************************
*/

void Update_Picture_Buffers()
{ 
  unsigned char ***tmp;
  unsigned char **tmp_y;
  int i;

#ifdef P_SUB_TYPE
  if(img->typeb==L2_IMG)
  {
	  return;
  }
  if(img->typeb==L3_IMG)
  {
	  tmp=ref_frm[1];
	  ref_frm[1]=ref_frm[0];
	  ref_frm[0]=tmp;

	  tmp_y=mref_frm[1];
	  mref_frm[1]=mref_frm[0];
	  mref_frm[0]=tmp_y;

	  //initial reference index, and for coming interpolation in mref[0]
	  for(i=0;i<3;i++)
	  {
		  mref[i] = mref_frm[i];
	  }
	  return;
  }
#endif

  //update integer pixel reference buffer
  tmp = ref_frm[1];       //ref_frm[ref_index][yuv][height][width] ref_index = 0,1  for P frame
  ref_frm[1] = ref_frm[0];    // ref_index = 0, backward reference for B frame; 1: forward reference for B frame
  ref_frm[0] = current_frame; // current_frame: current image under reconstruction
  current_frame = tmp;	

  //update luma 1/4 pixel reference buffer mref[ref_index][height][width] ref_index = 0,1 for P frame
  tmp_y = mref_frm[1];              // ref_index = 0, forward refernce for B frame ; 1: backward refernce for B frame
  mref_frm[1] = mref_frm[0];
  mref_frm[0] = tmp_y;

  //initial reference index, and for coming interpolation in mref[0]
  for(i=0;i<2;i++)
  {
    mref[i] = mref_frm[i];
  }
}

//edit start added by lzhang for AEC
void write_terminating_bit (unsigned char bit)
{
  DataPartition*          dataPart;
  EncodingEnvironmentPtr  eep_dp;

  //--- write non-slice termination symbol if the macroblock is not the first one in its slice ---
  dataPart = &(img->currentSlice->partArr[0]);
  dataPart->bitstream->write_flag = 1;
  eep_dp   = &(dataPart->ee_AEC);

  biari_encode_symbol_final(eep_dp, bit); 
#if TRACE
  fprintf (p_trace, "      aec_mb_stuffing_bit = %d\n",bit);
#endif

}

void free_slice()
{
  Slice * slice = img->currentSlice;
  if (slice != NULL)
  {
    if (slice->partArr != NULL)
      free(slice->partArr);
    if (input->symbol_mode == AEC)
    {
      delete_contexts_MotionInfo(slice->mot_ctx);
      delete_contexts_TextureInfo(slice->tex_ctx);
    }
    //free(img->currentSlice);
    free(slice);
  }
}